# pylint: disable=W0201, E1101
""" handle request for markdown pages """
import logging
import os
import tornado.web
import markdown
from .converter import Converter
from .access_control import UserMixin


LOGGER = logging.getLogger(__name__)
EMPTY_TOC = '<div class="toc">\n<ul></ul>\n</div>\n'


class SiteHandler(
    UserMixin, Converter, tornado.web.RequestHandler
):  # pylint: disable=W0223
    """ inline transform request for markdown pages """

    def initialize(self, docs, s3_loader):
        """ setup init properties """
        self.docs = docs
        self.meta = None
        self.nav = None
        self._s3_loader = s3_loader

    def create_template_loader(self, template_path):
        """ return s3 loader """
        if self._s3_loader:
            return self._s3_loader
        return super().create_template_loader(template_path)

    @property
    def has_toc(self):
        """ determin if toc is empty """
        return self.meta.toc != EMPTY_TOC

    def meta_value(self, name, default=None):
        """ return markdown meta value """
        return self.meta.Meta.get(name, [default])

    def one_meta_value(self, name, default=None):
        """ return markdown meta value """
        result = self.meta_value(name, default)
        return result[0] if result else None

    def load_nav(self, path):
        """ load nav section if it exist """
        folder = os.path.dirname(path)
        if folder:
            LOGGER.info(" -- folder: %s", folder)
            nav = os.path.join(self.docs, folder, "-nav.md")
            if os.path.isfile(nav):
                LOGGER.info(" -- nav: %s", nav)
                with open(nav, "r", encoding="utf-8") as file:
                    content = markdown.markdown(file.read())
                    self.nav = self.convert_images(content)

    def get(self, path):
        """ handle get """
        path = path if path else "index.html"
        file, ext = os.path.splitext(path)

        self.load_nav(path)

        doc = os.path.join(self.docs, f"{file}.md")
        # edit_path = os.path.join("/edit", f"{file}.md")
        edit_path = "/edit"
        with open(doc, "r", encoding="utf-8") as file:
            content = file.read()
            LOGGER.info(" -- ext: %s", ext)
            if ext == ".html":
                self.meta = self.markdown
                content = self.meta.convert(content)
                LOGGER.info(" -- meta: %s", self.meta.Meta)
                template = self.one_meta_value("template", "site")
                LOGGER.info(" -- tmpl: %s", template)
                self.render(
                    f"{template}_tmpl.html",
                    content=self.convert_images(content),
                    edit_path=edit_path,
                )
            else:
                self.write(self.convert_images(content))



# pylint: disable=R0914
""" make public folder of site """
import os
import shutil
from tornado import httpclient
from invoke import task
from .run import load_settings


@task
def publish(_, src, dst="public"):
    """ generate public site """
    settings = load_settings(src)

    port = settings.as_int("port", default="8080")
    http_client = httpclient.HTTPClient()

    public_path = dst
    if os.path.isdir(public_path):
        shutil.rmtree(public_path)

    pages = os.path.join(src, settings.get("pages_path", "pages"))
    for dirpath, _, filenames in os.walk(pages):
        for filename in filenames:
            if filename[0] not in ["-"]:
                page, ext = os.path.splitext(os.path.join(dirpath, filename))
                if ext == ".md":
                    web_page = os.path.relpath(page, pages)
                    web_page = f"{os.path.splitext(web_page)[0]}.html"
                    # print("page: ", f"{page}{ext}")
                    # print("\tweb: ",web_page)

                    public_page = os.path.join(public_path, web_page)
                    # print("\tpub: ", public_page)
                    url = f"http://localhost:{port}/{web_page}"
                    # print(url)
                    response = http_client.fetch(url)
                    folder = os.path.split(public_page)[0]
                    if folder and not os.path.exists(folder):
                        os.makedirs(folder)
                    with open(public_page, "wb") as file:
                        file.write(response.body)

    static = os.path.join(src, settings.get("static_path", "static"))
    public_static = os.path.join(public_path, "static")
    for dirpath, _, filenames in os.walk(static):
        if dirpath.endswith("/images"):
            continue
        for filename in filenames:
            fsrc = os.path.join(dirpath, filename)
            fdst = os.path.join(
                os.path.relpath(os.path.split(fsrc)[0], static), public_static
            )
            # print(fsrc, fdst)
            if fdst != ".":
                os.makedirs(fdst, exist_ok=True)
            shutil.copy(fsrc, fdst)



# pylint: disable=W0201
"""
    Simple mixin to provide utility methods
    to support User
"""
import logging
from json import dumps, loads
from tornado.web import RequestHandler, HTTPError

LOGGER = logging.getLogger(__name__)


class UserMixin:
    """ Support for user in tornado handlers """

    @property
    def cookie_name(self):
        """ return the cookie_name declared in application settings"""
        return self.settings.get("cookie_name")

    def get_current_user(self):
        """ return the current user from the cookie """
        if self.application.settings.get("debug") is True:
            token = self.request.headers.get("duck-token", None)
            if token:
                return token
        result = self.get_secure_cookie(self.cookie_name)
        if result:
            result = loads(result.decode("utf-8"))
        return result

    def set_current_user(self, value):
        """ put the current user in the cookie """
        if value:
            self.set_secure_cookie(self.cookie_name, dumps(value))
        else:
            self.clear_cookie(self.cookie_name)


class LoginHandler(UserMixin, RequestHandler):  # pylint: disable=W0223
    """
    Can be called as ajax from the
    websocket client to get the auth cookie
    into the headers.
    """

    def initialize(self, page=None, users=None):
        """ we're configured with a page """
        self.page = page if page else "login.html"
        self.register = None
        self.users = users

    def get_template_path(self):
        """ return app resource """
        return self.application.settings["duck_templates"]

    def login(self, username, password):
        """ return a user """
        user = None
        pwd = self.users.get(username)
        if pwd == password:
            user = username
            LOGGER.info("logged in: %s", user)
        return user

    def get(self, error=None, notice=None):
        """ render the form """
        email = self.get_argument("email", default=None)
        next_ = self.get_argument("next", "/")
        can_register = self.register is not None
        self.render(
            self.page,
            email=email,
            error=error,
            notice=notice,
            next=next_,
            can_register=can_register,
            app_name=self.settings.get("app_name", "duckdown"),
        )

    async def post(self):
        """ handle login post """
        try:
            email = self.get_argument("email", None)
            password = self.get_argument("password", None)
            submit = self.get_argument("submit", "login")
            if email is None or password is None:
                raise HTTPError(403, "email or password is None")

            user = None
            if submit == "login":
                user = self.login(email, password)
            elif self.register:
                user = self.register(email, password)  # pylint: disable=E1102
            if user:
                self.set_current_user(user)
                self.redirect(self.get_argument("next", "/"))
            else:
                raise Exception("email or password incorrect")
        except Exception as ex:  # pylint: disable=W0703
            LOGGER.exception(ex)
            self.get(error=str(ex))


class LogoutHandler(UserMixin, RequestHandler):  # pylint: disable=W0223
    """
    Removes the cookie from application settings
    and redirects.
    """

    def get(self):
        """ removes cookie and redirects to optional next argument """
        self.set_current_user(None)
        self.redirect(self.get_argument("next", "/"))



""" an s3 template loader """
import io
import os
import logging
from typing import Any
import boto3
from tornado.template import BaseLoader, Template

LOGGER = logging.getLogger(__name__)


class S3Loader(BaseLoader):
    """A template loader that loads from a single root directory."""

    def __init__(
        self,
        aws_access_key_id: str,
        aws_secret_access_key: str,
        bucket_name: str,
        folder="",
        **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)
        s3resource = boto3.resource(
            "s3",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )
        self.bucket = s3resource.Bucket(bucket_name)
        self.folder = folder

    def resolve_path(self, name, parent_path=None):
        """ resolve a path """
        if (
            parent_path
            and not parent_path.startswith("<")
            and not parent_path.startswith("/")
            and not name.startswith("/")
        ):
            current_path = os.path.join(self.folder, parent_path)
            file_dir = os.path.dirname(os.path.abspath(current_path))
            relative_path = os.path.abspath(os.path.join(file_dir, name))
            if relative_path.startswith(self.folder):
                name = relative_path[len(self.folder) + 1 :]
        return name

    def _create_template(self, name: str) -> Template:
        """ create a template from bucket/folder/name """
        key = "/".join([self.folder, name]) if self.folder else name
        LOGGER.info("template: %s", key)
        data = io.BytesIO()
        self.bucket.download_fileobj(key, data)
        template = Template(
            data.getvalue().decode("utf-8"), name=name, loader=self
        )
        return template



# pylint: disable=E1101
""" Mixin to convert to images to s3 """
import markdown
import pymdownx.emoji

EXTENSIONS = [
    "meta",
    "toc",
    "footnotes",
    "tables",
    "fenced_code",
    "attr_list",
    "def_list",
    "pymdownx.tilde",
    "pymdownx.tasklist",
    "pymdownx.emoji",
]

EXTENSIONS_CONFIG = {
    "pymdownx.emoji": {
        "emoji_index": pymdownx.emoji.twemoji,
        "emoji_generator": pymdownx.emoji.to_svg,
    }
}


class Converter:
    """ convert to remote if not debug """

    @property
    def markdown(self):
        """ returns a markdown instance """
        return markdown.Markdown(
            extensions=EXTENSIONS, extension_configs=EXTENSIONS_CONFIG
        )

    @property
    def img_path(self):
        """ return application img_path """
        return self.application.settings["img_path"]

    @property
    def local_images(self):
        """ return application local_images """
        return self.application.settings.get("local_images")

    def convert_images(self, value):
        """ use img_path """
        if self.local_images is True:
            return value
        return value.replace("/static/images/", self.img_path)



# pylint: disable=W0223
""" authenticated static files """
import logging
import tornado.web
from .access_control import UserMixin

LOGGER = logging.getLogger(__name__)


class AssetHandler(UserMixin, tornado.web.StaticFileHandler):
    """ you need to login to get these assets """

    @tornado.web.authenticated
    def prepare(self):
        """ nothing to be done """



# pylint: disable=W0201, W0223
""" browser and upload to bucket """
import os
import logging
import boto3
import tornado.web
from .base_handler import BaseHandler
from .json_utils import dumps
from .access_control import UserMixin

LOGGER = logging.getLogger(__name__)

TYPE_MAP = {
    ".svg": ("SVG", "image/svg+xml"),
    ".jpg": ("JPEG", "image/jpeg"),
    ".jpeg": ("JPEG", "image/jpeg"),
    ".png": ("PNG", "image/png"),
    ".gif": ("GIF", "image/gif"),
}


class S3Browser(UserMixin, BaseHandler):
    """ list contents of bucket """

    def initialize(
        self, aws_access_key_id, aws_secret_access_key, bucket_name, folder=""
    ):
        """ setup s3 bucket """
        self.name = bucket_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.folder = folder

    @property
    def bucket(self):
        """ return boto3 client """
        return boto3.client(
            "s3",
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
        )

    @property
    def static_path(self):
        """ return application static_path """
        return self.application.settings.get("static_path")

    @property
    def img_path(self):
        """ return application img_path """
        return self.application.settings.get("img_path")

    @property
    def local_images(self):
        """ return application local_images """
        return self.application.settings.get("local_images")

    def add(self, data=None, key=None, meta=None):
        """ adds data, returns path """
        if self.local_images:
            # move to static_path
            folder = self.folder if self.folder else ""
            if folder.endswith("/"):
                folder = folder[:-1]
            path = os.path.join(self.static_path, folder, key)
            LOGGER.info("adding %s", path)
            with open(path, "wb") as file:
                file.write(data)
            return str(path)

        # add to bucket
        key = self.folder + key if self.folder else key
        bucket = self.bucket
        bucket.put_object(
            ACL="public-read",
            Body=data,
            Bucket=self.name,
            Key=key,
            Metadata=meta,
        )
        return bucket.generate_presigned_url(
            "get_object", Params={"Bucket": self.name, "Key": key}
        )

    def list(self, prefix="", delimiter="/"):
        """ list the content of the bucket """
        folders = []
        files = []
        result = {
            "CommonPrefixes": folders,
            "Contents": files,
        }
        if self.local_images:
            # return to static_path
            path = os.path.join(self.static_path, self.folder, prefix)
            with os.scandir(path) as item:
                for entry in item:
                    if entry.is_file():
                        _, ext = os.path.splitext(entry.name)
                        if ext.lower() in TYPE_MAP:
                            file = f"{prefix}{entry.name}"
                            files.append({"Key": f"{file}"})
                    else:
                        folder = f"{prefix}{entry.name}/"
                        folders.append({"Prefix": f"{folder}"})
            return result

        # list bucket objects
        prefix = self.folder + prefix if self.folder else prefix
        items = self.bucket.list_objects(
            Bucket=self.name, Prefix=prefix, Delimiter=delimiter
        )
        starting = len(self.folder)
        for item in items.get("Contents", []):
            files.append({"Key": item["Key"][starting:]})
        for item in items.get("CommonPrefixes", []):
            folders.append({"Prefix": item["Prefix"][starting:]})
        return result

    @tornado.web.authenticated
    def get(self, prefix=None):
        """ returns the contents of bucket """
        if prefix is None:
            prefix = ""
        delimiter = self.get_argument("d", "/")
        self.set_header("Content-Type", "application/json")
        self.write(dumps(self.list(prefix, delimiter)))

    @tornado.web.authenticated
    def put(self, path=None):  # pylint: disable=W0613
        """ return the img_path """
        self.write({"img_path": self.img_path})

    @tornado.web.authenticated
    def post(self, path=None):
        """ handle the upload """
        result = []
        for key in self.request.files:
            for fileinfo in self.request.files[key]:
                fname = fileinfo["filename"]
                _, ext = os.path.splitext(fname)
                ftype, fmime = TYPE_MAP.get(ext.lower(), (None, None))
                if ftype is None:
                    raise Exception("File Type not accepted: {}".format(ftype))
                LOGGER.info("upload: %s", f"{path}{fname}")
                s3key = self.add(
                    fileinfo["body"],
                    key=f"{path}{fname}",
                    meta={"original_name": fname, "content-type": fmime},
                )
                result.append(s3key)
        self.write({"result": result})



# pylint: disable=W0223
""" We need access for development """
import tornado.web


class BaseHandler(tornado.web.RequestHandler):
    """ allow for development """

    def set_default_headers(self):
        """ allow access for development """
        if self.application.settings.get("debug") is True:
            self.set_header("Access-Control-Allow-Origin", "*")
            self.set_header("Access-Control-Allow-Headers", "duck-token")
            self.set_header(
                "Access-Control-Allow-Methods",
                "DELETE, PUT, POST, GET, OPTIONS",
            )

    def options(self, *args, **kwargs):
        """ allow dev request """
        if self.application.settings.get("debug") is True:
            self.set_status(204)
            self.finish()
        else:
            super().options(*args, **kwargs)



# pylint: disable=W0201
""" Handle request for index page """
from tornado.web import RequestHandler, StaticFileHandler, authenticated
from .access_control import UserMixin


class EditorHandler(UserMixin, RequestHandler):  # pylint: disable=W0223
    """ return index page """

    def initialize(self, manifest, page):
        """ setup production """
        self.manifest = manifest
        self.page = page

    @property
    def duck_path(self):
        """ return application duck_path """
        return self.application.settings.get("duck_path")

    @property
    def img_path(self):
        """ return application img_path """
        if self.application.settings.get("local_images"):
            path = self.application.settings.get("img_path", "")
            return f"/static/{path}"
        return self.application.settings.get("img_path")

    def get_template_path(self):
        """ return app resource """
        return self.application.settings["duck_templates"]

    def asset_url(self, path):
        """ return asset handler versioned url """
        return StaticFileHandler.make_static_url(
            {
                "static_url_prefix": "/edit/assets/",
                "static_path": self.application.settings["duck_assets"],
            },
            path,
        )

    @authenticated
    def get(self):
        """ handle get request """
        self.render(self.page)



""" handlers """
from .editor_handler import EditorHandler
from .site_handler import SiteHandler
from .dir_handler import DirHandler
from .mark_handler import MarkHandler
from .s3_browser import S3Browser
from .asset_handler import AssetHandler
from .access_control import LoginHandler, LogoutHandler



""" We want to manage a directory of files """
import os
import logging
import mimetypes
from .base_handler import BaseHandler
from .access_control import UserMixin

LOGGER = logging.getLogger(__name__)
mimetypes.add_type("text/markdown", ".md")


def scan_path(path):
    """ return the contents of a path """
    with os.scandir(path) as item:
        for entry in item:
            is_file = entry.is_file()
            size = entry.stat().st_size if is_file else None
            mime = mimetypes.guess_type(entry.path) if is_file else None
            if entry.name[0] != ".":
                yield {
                    "name": entry.name,
                    "file": is_file,
                    "size": size,
                    "type": mime,
                }


class DirHandler(UserMixin, BaseHandler):  # pylint: disable=W0223
    """ Manage a directory """

    def initialize(self, directory):
        """ setup directory """
        self.directory = directory  # pylint: disable=W0201

    def get(self, path=None):
        """ return the files and directories in path """
        path = os.path.join(self.directory, path) if path else self.directory
        if os.path.isfile(path):
            content_type, _ = mimetypes.guess_type(path)
            self.set_header("Content-Type", content_type)
            self.write(open(path, "r").read())
        else:
            self.write({"items": list(scan_path(path))})

    def put(self, path=None):
        """ handle the setting of file to path """
        LOGGER.info("saving %s", path)
        path = os.path.join(self.directory, path) if path else self.directory
        folder, _ = os.path.split(path)
        if folder and not os.path.exists(folder):
            LOGGER.info("making directory: %s", folder)
            os.makedirs(folder)
        with open(path, "wb") as file:
            file.write(self.request.body)
        self.write("saved")

    def delete(self, path=None):
        """ will remove a document """
        path = os.path.join(self.directory, path) if path else self.directory
        if os.path.isfile(path):
            os.unlink(path)
            self.write("deleted")



# pylint: disable=R0205, R0911, R0903
"""
    We want our own DateTimeEncoder
"""
import dataclasses
import datetime
import enum
import json
from collections.abc import Callable
from decimal import Decimal

try:
    from bson.objectid import ObjectId

except ImportError:

    class ObjectId:
        """ in case you're not using Mongodb """


class DateTimeEncoder(json.JSONEncoder):
    """
    Encodes datetimes and Decimals
    calls to_json on object if it has that method
    """

    def default(self, obj):  # pylint: disable=W0221,E0202
        """ check for our types """
        if hasattr(obj, "to_json") and isinstance(
            getattr(obj, "to_json"), Callable
        ):
            return obj.to_json()
        if dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)
        if isinstance(obj, enum.Enum):
            return obj.name
        if hasattr(obj, "isoformat"):
            return obj.isoformat().replace("T", " ")
        if isinstance(obj, Decimal):
            return float(obj)
        if isinstance(obj, ObjectId):
            return str(obj)
        return super().default(obj)


def loads(*args, **kwargs):
    """ calls json.loads """
    return json.loads(*args, **kwargs)


def dumps(obj, **kwargs):
    """ calls json.dumps using DateTimeEncoder """
    return json.dumps(obj, cls=DateTimeEncoder, **kwargs)


def type_from_json(type_, value):
    """provided with a string and type return the value as type"""
    result = value
    if value:
        if type_ == int:
            result = int(value)
        elif type_ == float:
            result = float(value)
        elif type_ == bool:
            result = str(value).lower() in ("yes", "true", "t", "1")
        elif type_ == Decimal:
            result = Decimal(value)
        elif type_ == datetime.datetime:
            result = (
                datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
                if value
                else None
            )
        elif type_ == datetime.date:
            result = (
                datetime.datetime.strptime(value, "%Y-%m-%d").date()
                if value
                else None
            )
        elif type_ == datetime.time:
            result = (
                datetime.datetime.strptime(value, "%H:%M:%S").time()
                if value
                else None
            )
    return result



""" Python install script """
from setuptools import setup, find_packages

try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

    class bdist_wheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.root_is_pure = True


except ImportError:
    bdist_wheel = None


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="duckdown",
    version="0.0.34",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/blueshed/duckdown/",
    packages=find_packages(exclude=["tests.*", "tests"]),
    package_data={
        "duckdown": ["assets/*","templates/*"],
        "duckdown.tool": ["data/*"]
    },
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "invoke",
        "convoke",
        "tornado",
        "markdown",
        "pymdown-extensions",
        "python-dotenv",
        "boto3"
    ],
    entry_points={"console_scripts": ["duckdown=duckdown.tool:program.run"]},
)



# pylint: disable=unused-variable, too-many-locals
""" application entry point """
import json
import logging
import os
import tornado.ioloop
import tornado.web
import tornado.log
import convoke
from dotenv import load_dotenv
from pkg_resources import resource_filename

from .handlers import EditorHandler
from .handlers import SiteHandler
from .handlers import DirHandler
from .handlers import MarkHandler
from .handlers import S3Browser
from .handlers import AssetHandler
from .handlers import LoginHandler, LogoutHandler
from .handlers.s3tmpl_loader import S3Loader

LOGGER = logging.getLogger(__name__)


def make_app_path(settings, name, default):
    """ return path relative to app """
    return os.path.join(
        settings.get("app_path", ""), settings.get(name, default)
    )


def make_app():
    """ make a tornado application """
    settings = convoke.get_settings("duckdown")
    debug = settings.as_bool("debug", default="False")
    production = settings.as_bool("production", default="True")

    if production is False:
        manifest = None
    else:
        LOGGER.info("loading client manifest")
        with open(
            resource_filename("duckdown", "assets/vue/manifest.json")
        ) as file:
            manifest = json.load(file)

    static_path = make_app_path(settings, "static_path", "static")
    template_path = make_app_path(settings, "template_path", "templates")
    pages_path = make_app_path(settings, "pages_path", "pages")
    users_path = make_app_path(settings, "users_path", "users.json")
    img_path = settings.get("img_path", "")

    tornado_settings = {
        "debug": debug,
        "production": production,
        "port": settings.as_int("port", default="8080"),
        "duck_users": users_path,
        "duck_path": "/edit/assets/",
        "duck_assets": resource_filename("duckdown", "assets"),
        "duck_templates": resource_filename("duckdown", "templates"),
        "static_path": settings.get("static", static_path),
        "template_path": settings.get("templates", template_path),
        "cookie_name": settings.get("cookie_name", "duckdown-cookie"),
        "cookie_secret": settings.get(
            "cookie_secret", "it was a dark and stormy duckdown"
        ),
        "login_url": settings.get("login_url", "/login"),
        "app_name": settings.get("app_name", "duckdown-app"),
        "local_images": settings.as_bool("local_images", default="False"),
        "img_path": img_path,
        "compress_response": True,
    }

    # load aws credentials
    image_bucket = {
        "bucket_name": settings.get("image_bucket", ""),
        "aws_access_key_id": settings.get("AWS_ACCESS_KEY_ID", ""),
        "aws_secret_access_key": settings.get("AWS_SECRET_ACCESS_KEY", ""),
        "folder": "images/",
    }

    LOGGER.info("settings:")
    for key, value in tornado_settings.items():
        LOGGER.info("\t%s: %r", key, value)

    s3_loader = S3Loader(
        **{
            "bucket_name": settings.get("image_bucket", ""),
            "aws_access_key_id": settings.get("AWS_ACCESS_KEY_ID", ""),
            "aws_secret_access_key": settings.get("AWS_SECRET_ACCESS_KEY", ""),
            "folder": "templates",
        }
    )

    routes = [
        (r"/login", LoginHandler, {"users": json.load(open(users_path))}),
        (r"/logout", LogoutHandler),
        (r"/browse/(.*)", S3Browser, image_bucket),
        (
            r"/edit/assets/(.*)",
            AssetHandler,
            {"path": tornado_settings["duck_assets"]},
        ),
        (r"/edit/mark/", MarkHandler),
        (r"/edit/pages/(.*)", DirHandler, {"directory": pages_path}),
        (r"/edit", EditorHandler, {"manifest": manifest, "page": "vue.html"}),
        (r"/(.*)", SiteHandler, {"docs": pages_path, "s3_loader": None}),
    ]

    if production is False:
        LOGGER.info("installing vue dev handler")
        routes.insert(
            0,
            (
                r"/src/(.*)",
                tornado.web.StaticFileHandler,
                {"path": "./client/src/"},
            ),
        )
    else:
        LOGGER.info("installing vue handler")
        _assets = resource_filename("duckdown", "assets/vue/")
        routes.insert(
            0,
            (
                r"/_assets/(.*)",
                tornado.web.StaticFileHandler,
                {"path": _assets},
            ),
        )

    return tornado.web.Application(routes, **tornado_settings)


def main():
    """ make an app and run it """
    load_dotenv(verbose=True)
    app = make_app()

    app.listen(app.settings["port"])
    LOGGER.info("listening on port: %s", app.settings["port"])

    if app.settings["debug"] is True:
        LOGGER.info("running in debug mode")

    ioloop = tornado.ioloop.IOLoop.current()
    try:
        ioloop.start()
    except KeyboardInterrupt:
        LOGGER.info("shutting down")
        ioloop.stop()


if __name__ == "__main__":
    tornado.log.enable_pretty_logging()
    main()



# pylint: disable=E1101
""" convert put to markdown """
import tornado.web
from .access_control import UserMixin
from .base_handler import BaseHandler
from .converter import Converter


class MarkHandler(UserMixin, Converter, BaseHandler):  # pylint: disable=W0223
    """ convert mardown put to json """

    @tornado.web.authenticated
    def put(self):
        """ handle put request """
        meta = self.markdown
        content = meta.convert(self.request.body.decode("utf-8"))
        content = self.convert_images(content)
        self.write({"content": content, "meta": meta.Meta, "toc": meta.toc})



""" task to create app template """
import os
from shutil import copyfile
from pkg_resources import resource_filename
from invoke import task


@task
def create(ctx, path, force=False):
    """ create a duckdown app at path """
    if os.path.exists(path):
        if force is False:
            print("already exists: ", path)
            return

        print("removing ", path)
        ctx.run(f"rm -rf {path}")

    print("creating ", path)
    sources = [
        ("", "config.ini"),
        ("", "users.json"),
        ("pages", "index.md"),
        ("static", "site.css"),
        ("static", "favicon.ico"),
        ("static/images", "logo.svg"),
        ("templates", "site_tmpl.html"),
    ]

    for folder in {folder for folder, _ in sources if folder}:
        os.makedirs(os.path.join(path, folder), exist_ok=True)

    for folder, file in sources:
        copyfile(
            resource_filename("duckdown.tool", f"data/{file}"),
            os.path.join(path, folder, file),
        )



"""
    This is a our invoke program
"""
import os
import sys
import tornado.log
from invoke import Program, Collection
from duckdown import VERSION
from .create import create
from .run import run
from .publish import publish

sys.path.insert(0, os.getcwd())

VERSION = "0.0.12"

_NAMESPACE_ = Collection()

_NAMESPACE_.add_task(create)
_NAMESPACE_.add_task(run)
_NAMESPACE_.add_task(publish)

tornado.log.enable_pretty_logging()

program = Program(
    version=VERSION, namespace=_NAMESPACE_
)  # pylint: disable=C0103



""" run duckdown app """
from pathlib import Path
import convoke
from invoke import task
from duckdown.main import main


def load_settings(path):
    """ load convoke settings from config.ini """
    settings = {"app_path": path}
    config = Path(f"{path}/config.ini")
    if config.exists():
        settings["config"] = config
    return convoke.get_settings("duckdown", **settings)


@task
def run(_, path):
    """ run app """
    load_settings(path)
    main()
